"""Tests for ExcelToAmplifyMigrator class"""

import pytest
import pandas as pd
from unittest.mock import MagicMock, patch
from migrator import ExcelToAmplifyMigrator


class TestToCamelCase:
    """Test to_camel_case static method"""

    def test_space_separated(self):
        """Test conversion with space-separated words"""
        assert ExcelToAmplifyMigrator.to_camel_case("user name") == "userName"
        assert ExcelToAmplifyMigrator.to_camel_case("User Name") == "userName"
        assert ExcelToAmplifyMigrator.to_camel_case("first name last name") == "firstNameLastName"

    def test_hyphen_separated(self):
        """Test conversion with hyphen-separated words"""
        assert ExcelToAmplifyMigrator.to_camel_case("user-name") == "userName"
        assert ExcelToAmplifyMigrator.to_camel_case("first-name") == "firstName"

    def test_underscore_separated(self):
        """Test conversion with underscore-separated words"""
        assert ExcelToAmplifyMigrator.to_camel_case("user_name") == "userName"
        assert ExcelToAmplifyMigrator.to_camel_case("first_name") == "firstName"

    def test_pascal_case(self):
        """Test conversion from PascalCase"""
        assert ExcelToAmplifyMigrator.to_camel_case("UserName") == "userName"
        assert ExcelToAmplifyMigrator.to_camel_case("FirstName") == "firstName"
        # Note: APIKey becomes aPIKey because each capital letter is treated as new word start
        assert ExcelToAmplifyMigrator.to_camel_case("APIKey") == "aPIKey"

    def test_already_camel_case(self):
        """Test that already camelCase strings remain unchanged"""
        assert ExcelToAmplifyMigrator.to_camel_case("firstName") == "firstName"
        assert ExcelToAmplifyMigrator.to_camel_case("userName") == "userName"

    def test_mixed_separators(self):
        """Test conversion with mixed separators"""
        assert ExcelToAmplifyMigrator.to_camel_case("user name-id") == "userNameId"
        assert ExcelToAmplifyMigrator.to_camel_case("first_name last-name") == "firstNameLastName"

    def test_single_word(self):
        """Test single word conversion"""
        assert ExcelToAmplifyMigrator.to_camel_case("user") == "user"
        assert ExcelToAmplifyMigrator.to_camel_case("User") == "user"

    def test_with_extra_spaces(self):
        """Test handling of extra spaces"""
        assert ExcelToAmplifyMigrator.to_camel_case("  user  name  ") == "userName"


class TestExcelToAmplifyMigratorInit:
    """Test ExcelToAmplifyMigrator initialization"""

    def test_initialization(self):
        """Test that migrator initializes correctly"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        assert migrator.excel_file_path == "test.xlsx"
        assert migrator.amplify_client is None
        assert migrator.model_field_parser is not None

    def test_init_client(self):
        """Test client initialization"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        with patch("migrator.AmplifyClient") as mock_client_class:
            mock_client = MagicMock()
            mock_client_class.return_value = mock_client

            migrator.init_client(
                api_endpoint="https://test.com",
                region="us-east-1",
                user_pool_id="pool-id",
                client_id="client-id",
                username="test@example.com",
            )

            # Verify AmplifyClient was created with correct parameters
            mock_client_class.assert_called_once_with(
                api_endpoint="https://test.com", user_pool_id="pool-id", region="us-east-1", client_id="client-id"
            )

            # Verify init_cognito_client was called
            mock_client.init_cognito_client.assert_called_once()


class TestTransformRowToRecord:
    """Test transform_row_to_record method"""

    def test_transforms_simple_row(self):
        """Test transforming a simple row"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Mock the parse_input method
        migrator.parse_input = MagicMock(side_effect=lambda row, field, _, fk_cache: row.get(field["name"]))

        row_dict = {"name": "John", "email": "john@example.com"}
        parsed_model_structure = {
            "fields": [{"name": "name", "is_required": True}, {"name": "email", "is_required": True}]
        }

        result = migrator.transform_row_to_record(row_dict, parsed_model_structure, {})

        assert result == {"name": "John", "email": "john@example.com"}

    def test_skips_none_values(self):
        """Test that None values are skipped"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Mock parse_input to return None for some fields
        def mock_parse_input(row, field, _, fk_cache):
            if field["name"] == "optional":
                return None
            return row.get(field["name"])

        migrator.parse_input = MagicMock(side_effect=mock_parse_input)

        row_dict = {"name": "John", "optional": None}
        parsed_model_structure = {
            "fields": [{"name": "name", "is_required": True}, {"name": "optional", "is_required": False}]
        }

        result = migrator.transform_row_to_record(row_dict, parsed_model_structure, {})

        assert result == {"name": "John"}
        assert "optional" not in result


class TestTransformRowsToRecords:
    """Test transform_rows_to_records method"""

    def test_transforms_dataframe(self):
        """Test transforming entire DataFrame"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Initialize instance state
        migrator.failed_records_by_sheet = {}
        migrator._current_sheet = "TestSheet"
        migrator.failed_records_by_sheet["TestSheet"] = []

        # Mock amplify_client
        migrator.amplify_client = MagicMock()
        migrator.amplify_client.get_primary_field_name.return_value = ("name", False, "String")

        # Mock transform_row_to_record
        migrator.transform_row_to_record = MagicMock(
            side_effect=[{"name": "John", "email": "john@example.com"}, {"name": "Jane", "email": "jane@example.com"}]
        )

        df = pd.DataFrame({"Name": ["John", "Jane"], "Email": ["john@example.com", "jane@example.com"]})

        parsed_model_structure = {"fields": []}

        records, row_dict_by_primary = migrator.transform_rows_to_records(df, parsed_model_structure, "TestSheet")

        assert len(records) == 2
        assert records[0]["name"] == "John"
        assert records[1]["name"] == "Jane"
        assert len(migrator.failed_records_by_sheet["TestSheet"]) == 0

    def test_converts_column_names_to_camel_case(self):
        """Test that DataFrame columns are converted to camelCase"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Initialize instance state
        migrator.failed_records_by_sheet = {}
        migrator._current_sheet = "TestSheet"
        migrator.failed_records_by_sheet["TestSheet"] = []

        # Mock amplify_client
        migrator.amplify_client = MagicMock()
        migrator.amplify_client.get_primary_field_name.return_value = ("userName", False, "String")

        # Mock transform_row_to_record to capture the row
        captured_rows = []

        def capture_row(row, _, fk_cache):
            captured_rows.append(row)
            return {"test": "value"}

        migrator.transform_row_to_record = MagicMock(side_effect=capture_row)

        df = pd.DataFrame({"User Name": ["John"], "Email Address": ["john@example.com"]})

        parsed_model_structure = {"fields": []}

        records, row_dict_by_primary = migrator.transform_rows_to_records(df, parsed_model_structure, "TestSheet")

        # Check that columns were converted to camelCase
        assert "userName" in captured_rows[0]
        assert "emailAddress" in captured_rows[0]

    def test_handles_errors_in_rows(self):
        """Test that errors in individual rows don't stop processing"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Initialize instance state
        migrator.failed_records_by_sheet = {}
        migrator._current_sheet = "TestSheet"
        migrator.failed_records_by_sheet["TestSheet"] = []

        # Mock amplify_client
        migrator.amplify_client = MagicMock()
        migrator.amplify_client.get_primary_field_name.return_value = ("name", False, "String")

        # Mock to raise error for second row
        call_count = [0]

        def mock_transform(row, _, fk_cache):
            call_count[0] += 1
            if call_count[0] == 2:  # Second call
                raise ValueError("Test error")
            return {"name": row["name"]}

        migrator.transform_row_to_record = MagicMock(side_effect=mock_transform)

        df = pd.DataFrame({"name": ["John", "Jane", "Bob"]})

        parsed_model_structure = {"fields": []}

        records, row_dict_by_primary = migrator.transform_rows_to_records(df, parsed_model_structure, "TestSheet")

        # Should have 2 records (skipped the one with error)
        assert len(records) == 2
        assert records[0]["name"] == "John"
        assert records[1]["name"] == "Bob"

        # Should have 1 failed parsing record in instance state
        failed_parsing = migrator.failed_records_by_sheet["TestSheet"]
        assert len(failed_parsing) == 1
        assert failed_parsing[0]["primary_field"] == "name"
        assert failed_parsing[0]["primary_field_value"] == "Jane"
        assert "Parsing error: Test error" in failed_parsing[0]["error"]

    def test_tracks_multiple_parsing_failures(self):
        """Test that multiple parsing failures are tracked"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Initialize instance state
        migrator.failed_records_by_sheet = {}
        migrator._current_sheet = "TestSheet"
        migrator.failed_records_by_sheet["TestSheet"] = []

        # Mock amplify_client
        migrator.amplify_client = MagicMock()
        migrator.amplify_client.get_primary_field_name.return_value = ("name", False, "String")

        # Mock to raise error for rows 2 and 4
        call_count = [0]

        def mock_transform(row, _, fk_cache):
            call_count[0] += 1
            if call_count[0] in [2, 4]:
                raise ValueError(f"Error in row {call_count[0]}")
            return {"name": row["name"]}

        migrator.transform_row_to_record = MagicMock(side_effect=mock_transform)

        df = pd.DataFrame({"name": ["Alice", "Bob", "Charlie", "David", "Eve"]})

        parsed_model_structure = {"fields": []}

        records, row_dict_by_primary = migrator.transform_rows_to_records(df, parsed_model_structure, "TestSheet")

        # Should have 3 successful records
        assert len(records) == 3
        assert records[0]["name"] == "Alice"
        assert records[1]["name"] == "Charlie"
        assert records[2]["name"] == "Eve"

        # Should have 2 failed parsing records in instance state
        failed_parsing = migrator.failed_records_by_sheet["TestSheet"]
        assert len(failed_parsing) == 2
        assert failed_parsing[0]["primary_field_value"] == "Bob"
        assert failed_parsing[1]["primary_field_value"] == "David"

    def test_uses_row_number_when_primary_field_missing(self):
        """Test that row number is used when primary field value is missing"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Initialize instance state
        migrator.failed_records_by_sheet = {}
        migrator._current_sheet = "TestSheet"
        migrator.failed_records_by_sheet["TestSheet"] = []

        # Mock amplify_client
        migrator.amplify_client = MagicMock()
        migrator.amplify_client.get_primary_field_name.return_value = ("name", False, "String")

        def mock_transform(row, _, fk_cache):
            raise ValueError("Missing required field")

        migrator.transform_row_to_record = MagicMock(side_effect=mock_transform)

        # DataFrame without the primary field
        df = pd.DataFrame({"otherField": ["value1"]})

        parsed_model_structure = {"fields": []}

        records, row_dict_by_primary = migrator.transform_rows_to_records(df, parsed_model_structure, "TestSheet")

        # Should have 1 failed parsing record with row number as fallback
        failed_parsing = migrator.failed_records_by_sheet["TestSheet"]
        assert len(failed_parsing) == 1
        assert failed_parsing[0]["primary_field_value"] == "Row 1"

    def test_stores_original_row_in_failure_records(self):
        """Test that original row data is stored when parsing fails"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Initialize instance state
        migrator.failed_records_by_sheet = {}
        migrator._current_sheet = "TestSheet"
        migrator.failed_records_by_sheet["TestSheet"] = []

        # Mock amplify_client
        migrator.amplify_client = MagicMock()
        migrator.amplify_client.get_primary_field_name.return_value = ("name", False, "String")

        def mock_transform(row, _, fk_cache):
            raise ValueError("Test error")

        migrator.transform_row_to_record = MagicMock(side_effect=mock_transform)

        df = pd.DataFrame({"name": ["John"], "email": ["john@example.com"], "age": [30]})

        parsed_model_structure = {"fields": []}

        records, row_dict_by_primary = migrator.transform_rows_to_records(df, parsed_model_structure, "TestSheet")

        # Verify original row is stored
        failed_parsing = migrator.failed_records_by_sheet["TestSheet"]
        assert len(failed_parsing) == 1
        assert "original_row" in failed_parsing[0]
        assert failed_parsing[0]["original_row"]["name"] == "John"
        assert failed_parsing[0]["original_row"]["email"] == "john@example.com"
        assert failed_parsing[0]["original_row"]["age"] == 30

    def test_builds_row_dict_by_primary_mapping(self):
        """Test that row_dict_by_primary mapping is correctly built"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Initialize instance state
        migrator.failed_records_by_sheet = {}
        migrator._current_sheet = "TestSheet"
        migrator.failed_records_by_sheet["TestSheet"] = []

        # Mock amplify_client
        migrator.amplify_client = MagicMock()
        migrator.amplify_client.get_primary_field_name.return_value = ("name", False, "String")

        # Mock transform_row_to_record
        migrator.transform_row_to_record = MagicMock(
            side_effect=[{"name": "John", "email": "john@example.com"}, {"name": "Jane", "email": "jane@example.com"}]
        )

        df = pd.DataFrame({"Name": ["John", "Jane"], "Email": ["john@example.com", "jane@example.com"]})

        parsed_model_structure = {"fields": []}

        records, row_dict_by_primary = migrator.transform_rows_to_records(df, parsed_model_structure, "TestSheet")

        # Verify row_dict_by_primary has correct mappings
        assert "John" in row_dict_by_primary
        assert "Jane" in row_dict_by_primary
        assert row_dict_by_primary["John"]["name"] == "John"
        assert row_dict_by_primary["John"]["email"] == "john@example.com"
        assert row_dict_by_primary["Jane"]["name"] == "Jane"
        assert row_dict_by_primary["Jane"]["email"] == "jane@example.com"


class TestWriteFailedRecordsToExcel:
    """Test _write_failed_records_to_excel method"""

    def test_writes_excel_with_original_row_data(self, tmp_path):
        """Test that Excel file is created with original row data"""
        excel_file = tmp_path / "test.xlsx"
        migrator = ExcelToAmplifyMigrator(str(excel_file))

        # Initialize instance state with failures
        migrator.failed_records_by_sheet = {
            "Sheet1": [
                {
                    "primary_field": "name",
                    "primary_field_value": "John",
                    "error": "Parsing error: Required field missing",
                    "original_row": {"name": "John", "email": "john@example.com", "age": 30},
                },
                {
                    "primary_field": "name",
                    "primary_field_value": "Jane",
                    "error": "Upload error: GraphQL failed",
                    "original_row": {"name": "Jane", "email": "jane@example.com", "age": 25},
                },
            ]
        }

        # Write failed records to Excel
        output_path = migrator._write_failed_records_to_excel()

        # Verify file was created with timestamp
        assert output_path is not None
        assert "test_failed_records_" in output_path
        assert output_path.endswith(".xlsx")

        # Read and verify Excel contents
        df = pd.read_excel(output_path, sheet_name="Sheet1")

        # Verify columns exist
        assert "name" in df.columns
        assert "email" in df.columns
        assert "age" in df.columns
        assert "ERROR" in df.columns

        # Verify data
        assert len(df) == 2
        assert df.loc[0, "name"] == "John"
        assert df.loc[0, "email"] == "john@example.com"
        assert df.loc[0, "age"] == 30
        assert "Required field missing" in df.loc[0, "ERROR"]

        assert df.loc[1, "name"] == "Jane"
        assert df.loc[1, "email"] == "jane@example.com"
        assert df.loc[1, "age"] == 25
        assert "GraphQL failed" in df.loc[1, "ERROR"]

    def test_returns_none_when_no_failures(self, tmp_path):
        """Test that None is returned when there are no failures"""
        excel_file = tmp_path / "test.xlsx"
        migrator = ExcelToAmplifyMigrator(str(excel_file))

        # Initialize with empty failures
        migrator.failed_records_by_sheet = {}

        output_path = migrator._write_failed_records_to_excel()

        assert output_path is None

    def test_returns_none_when_all_sheets_have_empty_failures(self, tmp_path):
        """Test that None is returned when all sheets have empty failure lists"""
        excel_file = tmp_path / "test.xlsx"
        migrator = ExcelToAmplifyMigrator(str(excel_file))

        # Initialize with empty failure lists
        migrator.failed_records_by_sheet = {"Sheet1": [], "Sheet2": []}

        output_path = migrator._write_failed_records_to_excel()

        assert output_path is None

    def test_handles_multiple_sheets(self, tmp_path):
        """Test that multiple sheets are written correctly"""
        excel_file = tmp_path / "test.xlsx"
        migrator = ExcelToAmplifyMigrator(str(excel_file))

        # Initialize with failures in multiple sheets
        migrator.failed_records_by_sheet = {
            "Users": [
                {
                    "primary_field": "name",
                    "primary_field_value": "John",
                    "error": "Error 1",
                    "row_number": 1,
                    "original_row": {"name": "John"},
                }
            ],
            "Posts": [
                {
                    "primary_field": "title",
                    "primary_field_value": "Post1",
                    "error": "Error 2",
                    "row_number": 5,
                    "original_row": {"title": "Post1"},
                }
            ],
        }

        output_path = migrator._write_failed_records_to_excel()

        # Verify both sheets were created
        assert output_path is not None

        df_users = pd.read_excel(output_path, sheet_name="Users")
        df_posts = pd.read_excel(output_path, sheet_name="Posts")

        assert len(df_users) == 1
        assert df_users.loc[0, "name"] == "John"

        assert len(df_posts) == 1
        assert df_posts.loc[0, "title"] == "Post1"


class TestReadExcel:
    """Test read_excel method"""

    def test_reads_excel_file(self, tmp_path):
        """Test reading Excel file"""
        # Create a test Excel file
        excel_file = tmp_path / "test.xlsx"
        df = pd.DataFrame({"Name": ["John", "Jane"], "Age": [30, 25]})
        df.to_excel(excel_file, sheet_name="Users", index=False)

        migrator = ExcelToAmplifyMigrator(str(excel_file))

        sheets = migrator.read_excel()

        assert "Users" in sheets
        assert len(sheets["Users"]) == 2
        assert list(sheets["Users"]["Name"]) == ["John", "Jane"]

    def test_reads_multiple_sheets(self, tmp_path):
        """Test reading Excel file with multiple sheets"""
        excel_file = tmp_path / "test.xlsx"

        with pd.ExcelWriter(excel_file) as writer:
            df1 = pd.DataFrame({"Name": ["John"]})
            df2 = pd.DataFrame({"Title": ["Post1"]})
            df1.to_excel(writer, sheet_name="Users", index=False)
            df2.to_excel(writer, sheet_name="Posts", index=False)

        migrator = ExcelToAmplifyMigrator(str(excel_file))

        sheets = migrator.read_excel()

        assert len(sheets) == 2
        assert "Users" in sheets
        assert "Posts" in sheets


class TestParseInputWithRelationships:
    """Test parse_input method with belongsTo relationships"""

    def test_uses_related_model_when_present(self):
        """Test that parse_input uses related_model from field metadata"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Mock amplify_client
        mock_client = MagicMock()
        mock_client.get_record.return_value = {"id": "reporter-123"}
        migrator.amplify_client = mock_client

        row_dict = {"photographer": "John Doe"}
        field = {
            "name": "photographerId",
            "is_id": True,
            "is_required": True,
            "related_model": "Reporter",  # This should be used instead of inferring "Photographer"
        }
        parsed_model_structure = {"fields": []}
        fk_lookup_cache = {"Reporter": {"lookup": {"John Doe": "reporter-123"}}}

        result = migrator.parse_input(row_dict, field, parsed_model_structure, fk_lookup_cache)

        # Should use the lookup cache
        assert result == "reporter-123"

    def test_falls_back_to_field_name_inference(self):
        """Test that parse_input falls back to field name inference when related_model absent"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Mock amplify_client
        mock_client = MagicMock()
        mock_client.get_record.return_value = {"id": "user-456"}
        migrator.amplify_client = mock_client

        row_dict = {"author": "Jane Smith"}
        field = {
            "name": "authorId",
            "is_id": True,
            "is_required": False,
            # No related_model property - should infer "Author" from "authorId"
        }
        parsed_model_structure = {"fields": []}
        fk_lookup_cache = {"Author": {"lookup": {"Jane Smith": "user-456"}}}

        result = migrator.parse_input(row_dict, field, parsed_model_structure, fk_lookup_cache)

        # Should use the lookup cache
        assert result == "user-456"

    def test_handles_missing_related_record(self):
        """Test handling when related record is not found"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Mock amplify_client to return None
        mock_client = MagicMock()
        mock_client.get_record.return_value = None
        migrator.amplify_client = mock_client

        row_dict = {"photographer": "Unknown Person"}
        field = {"name": "photographerId", "is_id": True, "is_required": True, "related_model": "Reporter"}
        parsed_model_structure = {"fields": []}
        fk_lookup_cache = {"Reporter": {"lookup": {}}}  # Empty lookup - record not found

        with pytest.raises(ValueError, match="Reporter: Unknown Person does not exist"):
            migrator.parse_input(row_dict, field, parsed_model_structure, fk_lookup_cache)

    def test_handles_related_record_with_null_id(self):
        """Test handling when related record exists but has null ID"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Mock amplify_client to return record with null id
        mock_client = MagicMock()
        mock_client.get_record.return_value = {"id": None}
        migrator.amplify_client = mock_client

        row_dict = {"photographer": "John Doe"}
        field = {"name": "photographerId", "is_id": True, "is_required": True, "related_model": "Reporter"}
        parsed_model_structure = {"fields": []}
        fk_lookup_cache = {"Reporter": {"lookup": {}}}  # Empty lookup

        # Should raise ValueError because field is required and ID is not in cache
        with pytest.raises(ValueError, match="Reporter: John Doe does not exist"):
            migrator.parse_input(row_dict, field, parsed_model_structure, fk_lookup_cache)

    def test_handles_optional_id_field_with_null(self):
        """Test handling optional ID field when related record ID is null"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Mock amplify_client to return record with null id
        mock_client = MagicMock()
        mock_client.get_record.return_value = {"id": None}
        migrator.amplify_client = mock_client

        row_dict = {"photographer": "John Doe"}
        field = {
            "name": "photographerId",
            "is_id": True,
            "is_required": False,  # Optional field
            "related_model": "Reporter",
        }
        parsed_model_structure = {"fields": []}
        fk_lookup_cache = {"Reporter": {"lookup": {}}}  # Empty lookup

        result = migrator.parse_input(row_dict, field, parsed_model_structure, fk_lookup_cache)

        # Should return None without raising error for optional field
        assert result is None

    def test_integration_with_real_scenario(self):
        """Test full integration with Story -> Reporter relationship"""
        migrator = ExcelToAmplifyMigrator("test.xlsx")

        # Mock amplify_client
        mock_client = MagicMock()
        mock_client.get_record.return_value = {"id": "reporter-abc-123"}
        migrator.amplify_client = mock_client

        # Simulate Excel row
        row_dict = {"title": "Breaking News", "photographer": "Alice Johnson", "content": "Story content here"}

        # Field definition with related_model
        photographer_id_field = {
            "name": "photographerId",
            "is_id": True,
            "is_required": True,
            "related_model": "Reporter",  # Comes from: photographer: a.belongsTo("Reporter", "photographerId")
        }

        parsed_model_structure = {"fields": []}
        fk_lookup_cache = {"Reporter": {"lookup": {"Alice Johnson": "reporter-abc-123"}}}

        result = migrator.parse_input(row_dict, photographer_id_field, parsed_model_structure, fk_lookup_cache)

        # Should use the lookup cache
        assert result == "reporter-abc-123"



"""Tests for CLI commands"""

import json
import sys
import pytest
from pathlib import Path
from unittest.mock import patch, MagicMock
from migrator import (
    cmd_show,
    cmd_config,
    cmd_migrate,
)
from amplify_excel_migrator.core import ConfigManager


@pytest.fixture
def sample_config():
    """Sample configuration for testing"""
    return {
        "excel_path": "test_data.xlsx",
        "api_endpoint": "https://test.appsync-api.us-east-1.amazonaws.com/graphql",
        "region": "us-east-1",
        "user_pool_id": "us-east-1_testpool",
        "client_id": "test-client-id",
        "username": "test@example.com",
    }


@pytest.fixture
def mock_config_manager(tmp_path):
    """Create a mock ConfigManager for testing"""
    test_config_file = tmp_path / "config.json"

    def init_mock(self, config_path=None):
        self.config_path = test_config_file
        self._config = {}

    return init_mock


class TestCmdShow:
    """Test 'show' command"""

    def test_show_with_no_config(self, capsys, mock_config_manager):
        """Test show command with no config file"""
        with patch.object(ConfigManager, "__init__", mock_config_manager):
            cmd_show()

        captured = capsys.readouterr()
        assert "❌ No configuration found!" in captured.out
        assert "amplify-migrator config" in captured.out

    def test_show_with_existing_config(self, capsys, tmp_path, sample_config):
        """Test show command with existing config"""
        test_config_file = tmp_path / "config.json"
        test_config_file.parent.mkdir(parents=True, exist_ok=True)

        with open(test_config_file, "w") as f:
            json.dump(sample_config, f)

        def init_mock(self, config_path=None):
            self.config_path = test_config_file
            self._config = {}

        with patch.object(ConfigManager, "__init__", init_mock):
            cmd_show()

        captured = capsys.readouterr()
        assert "test_data.xlsx" in captured.out
        assert "test@example.com" in captured.out
        assert "us-east-1" in captured.out
        assert "test-client-id" in captured.out

    def test_show_displays_config_location(self, capsys, tmp_path, sample_config):
        """Test that show command displays config file location"""
        test_config_file = tmp_path / "config.json"
        test_config_file.parent.mkdir(parents=True, exist_ok=True)

        with open(test_config_file, "w") as f:
            json.dump(sample_config, f)

        def init_mock(self, config_path=None):
            self.config_path = test_config_file
            self._config = {}

        with patch.object(ConfigManager, "__init__", init_mock):
            cmd_show()

        captured = capsys.readouterr()
        assert str(test_config_file) in captured.out


class TestCmdConfig:
    """Test 'config' command"""

    def test_config_prompts_for_all_values(self, tmp_path):
        """Test that config command prompts for all required values"""
        test_config_file = tmp_path / "config.json"

        def init_mock(self, config_path=None):
            self.config_path = test_config_file
            self._config = {}

        # Mock all input prompts
        inputs = [
            "test.xlsx",
            "https://test.appsync-api.us-east-1.amazonaws.com/graphql",
            "us-east-1",
            "us-east-1_test",
            "test-client",
            "admin@test.com",
        ]

        with patch.object(ConfigManager, "__init__", init_mock):
            with patch("builtins.input", side_effect=inputs):
                cmd_config()

        # Verify config was saved
        assert test_config_file.exists()
        with open(test_config_file) as f:
            saved_config = json.load(f)
            assert saved_config["excel_path"] == "test.xlsx"
            assert saved_config["username"] == "admin@test.com"

    def test_config_saves_to_correct_location(self, capsys, tmp_path):
        """Test that config is saved to the correct location"""
        test_config_file = tmp_path / "config.json"

        def init_mock(self, config_path=None):
            self.config_path = test_config_file
            self._config = {}

        inputs = ["test.xlsx", "https://test.com", "us-east-1", "pool", "client", "user"]

        with patch.object(ConfigManager, "__init__", init_mock):
            with patch("builtins.input", side_effect=inputs):
                cmd_config()

        captured = capsys.readouterr()
        assert "✅ Configuration saved successfully!" in captured.out
        assert "amplify-migrator migrate" in captured.out

    def test_config_uses_cached_values_as_defaults(self, tmp_path, sample_config):
        """Test that config command shows cached values as defaults"""
        test_config_file = tmp_path / "config.json"
        test_config_file.parent.mkdir(parents=True, exist_ok=True)

        with open(test_config_file, "w") as f:
            json.dump(sample_config, f)

        def init_mock(self, config_path=None):
            self.config_path = test_config_file
            self._config = {}

        # Press enter to accept all defaults (from cached config)
        inputs = ["", "", "", "", "", ""]  # Empty strings use cached values

        with patch.object(ConfigManager, "__init__", init_mock):
            with patch("builtins.input", side_effect=inputs):
                cmd_config()

        with open(test_config_file) as f:
            saved_config = json.load(f)
            assert saved_config["excel_path"] == "test_data.xlsx"
            assert saved_config["region"] == "us-east-1"


class TestCmdMigrate:
    """Test 'migrate' command"""

    def test_migrate_fails_without_config(self, capsys, tmp_path):
        """Test that migrate command fails when no config exists"""
        test_config_file = tmp_path / "config.json"

        def init_mock(self, config_path=None):
            self.config_path = test_config_file
            self._config = {}

        with patch.object(ConfigManager, "__init__", init_mock):
            with pytest.raises(SystemExit) as exc_info:
                cmd_migrate()

        assert exc_info.value.code == 1

        captured = capsys.readouterr()
        assert "❌ No configuration found!" in captured.out
        assert "amplify-migrator config" in captured.out

    def test_migrate_uses_cached_config(self, tmp_path, sample_config):
        """Test that migrate command uses cached configuration"""
        test_config_file = tmp_path / "config.json"
        test_config_file.parent.mkdir(parents=True, exist_ok=True)

        with open(test_config_file, "w") as f:
            json.dump(sample_config, f)

        def init_mock(self, config_path=None):
            self.config_path = test_config_file
            self._config = {}

        # Mock the entire migration process
        with patch.object(ConfigManager, "__init__", init_mock):
            with patch("migrator.ExcelToAmplifyMigrator") as mock_migrator_class:
                with patch("amplify_excel_migrator.core.config.getpass", return_value="password123"):
                    mock_instance = MagicMock()
                    mock_instance.authenticate.return_value = True
                    mock_migrator_class.return_value = mock_instance

                    cmd_migrate()

                    # Verify migrator was initialized with cached values
                    mock_migrator_class.assert_called_once_with("test_data.xlsx")
                    mock_instance.init_client.assert_called_once()

                    # Check that init_client was called with correct parameters
                    call_args = mock_instance.init_client.call_args
                    assert call_args[0][0] == "https://test.appsync-api.us-east-1.amazonaws.com/graphql"
                    assert call_args[0][1] == "us-east-1"
                    assert call_args[0][2] == "us-east-1_testpool"

    def test_migrate_prompts_for_password(self, tmp_path, sample_config):
        """Test that migrate command always prompts for password"""
        test_config_file = tmp_path / "config.json"
        test_config_file.parent.mkdir(parents=True, exist_ok=True)

        with open(test_config_file, "w") as f:
            json.dump(sample_config, f)

        def init_mock(self, config_path=None):
            self.config_path = test_config_file
            self._config = {}

        with patch.object(ConfigManager, "__init__", init_mock):
            with patch("migrator.ExcelToAmplifyMigrator") as mock_migrator_class:
                with patch(
                    "amplify_excel_migrator.core.config.getpass", return_value="secret_password"
                ) as mock_getpass:
                    mock_instance = MagicMock()
                    mock_instance.authenticate.return_value = True
                    mock_migrator_class.return_value = mock_instance

                    cmd_migrate()

                    # Verify getpass was called (for password prompt)
                    mock_getpass.assert_called()

    def test_migrate_stops_if_authentication_fails(self, tmp_path, sample_config):
        """Test that migrate stops if authentication fails"""
        test_config_file = tmp_path / "config.json"
        test_config_file.parent.mkdir(parents=True, exist_ok=True)

        with open(test_config_file, "w") as f:
            json.dump(sample_config, f)

        def init_mock(self, config_path=None):
            self.config_path = test_config_file
            self._config = {}

        with patch.object(ConfigManager, "__init__", init_mock):
            with patch("migrator.ExcelToAmplifyMigrator") as mock_migrator_class:
                with patch("amplify_excel_migrator.core.config.getpass", return_value="wrong_password"):
                    mock_instance = MagicMock()
                    mock_instance.authenticate.return_value = False  # Authentication fails
                    mock_migrator_class.return_value = mock_instance

                    cmd_migrate()

                    # Verify run() was NOT called
                    mock_instance.run.assert_not_called()



"""Tests for MutationBuilder class"""

import pytest
from amplify_excel_migrator.graphql import MutationBuilder


class TestBuildCreateMutation:
    """Test MutationBuilder.build_create_mutation() method"""

    def test_create_mutation_basic(self):
        """Test basic create mutation"""
        mutation = MutationBuilder.build_create_mutation("User", return_fields=["id", "name", "email"])

        assert "mutation CreateUser" in mutation
        assert "$input: CreateUserInput!" in mutation
        assert "createUser(input: $input)" in mutation
        assert "id" in mutation
        assert "name" in mutation
        assert "email" in mutation

    def test_create_mutation_default_fields(self):
        """Test create mutation with default fields (only id)"""
        mutation = MutationBuilder.build_create_mutation("Product")

        assert "mutation CreateProduct" in mutation
        assert "$input: CreateProductInput!" in mutation
        assert "createProduct(input: $input)" in mutation
        assert "id" in mutation

    def test_create_mutation_single_field(self):
        """Test create mutation with single return field"""
        mutation = MutationBuilder.build_create_mutation("Post", return_fields=["title"])

        assert "mutation CreatePost" in mutation
        assert "title" in mutation

    def test_create_mutation_many_fields(self):
        """Test create mutation with many return fields"""
        fields = ["id", "title", "content", "authorID", "createdAt", "updatedAt"]
        mutation = MutationBuilder.build_create_mutation("Article", return_fields=fields)

        for field in fields:
            assert field in mutation

    def test_create_mutation_different_models(self):
        """Test create mutation with different model names"""
        models = ["User", "Product", "Order", "Comment", "BlogPost"]

        for model in models:
            mutation = MutationBuilder.build_create_mutation(model)
            assert f"mutation Create{model}" in mutation
            assert f"$input: Create{model}Input!" in mutation
            assert f"create{model}(input: $input)" in mutation


class TestBuildUpdateMutation:
    """Test MutationBuilder.build_update_mutation() method"""

    def test_update_mutation_basic(self):
        """Test basic update mutation"""
        mutation = MutationBuilder.build_update_mutation("User", return_fields=["id", "name", "email"])

        assert "mutation UpdateUser" in mutation
        assert "$input: UpdateUserInput!" in mutation
        assert "updateUser(input: $input)" in mutation
        assert "id" in mutation
        assert "name" in mutation
        assert "email" in mutation

    def test_update_mutation_default_fields(self):
        """Test update mutation with default fields (only id)"""
        mutation = MutationBuilder.build_update_mutation("Product")

        assert "mutation UpdateProduct" in mutation
        assert "$input: UpdateProductInput!" in mutation
        assert "updateProduct(input: $input)" in mutation
        assert "id" in mutation

    def test_update_mutation_single_field(self):
        """Test update mutation with single return field"""
        mutation = MutationBuilder.build_update_mutation("Post", return_fields=["updatedAt"])

        assert "mutation UpdatePost" in mutation
        assert "updatedAt" in mutation

    def test_update_mutation_many_fields(self):
        """Test update mutation with many return fields"""
        fields = ["id", "title", "content", "status", "updatedAt"]
        mutation = MutationBuilder.build_update_mutation("Article", return_fields=fields)

        for field in fields:
            assert field in mutation


class TestBuildDeleteMutation:
    """Test MutationBuilder.build_delete_mutation() method"""

    def test_delete_mutation_basic(self):
        """Test basic delete mutation"""
        mutation = MutationBuilder.build_delete_mutation("User", return_fields=["id", "name"])

        assert "mutation DeleteUser" in mutation
        assert "$input: DeleteUserInput!" in mutation
        assert "deleteUser(input: $input)" in mutation
        assert "id" in mutation
        assert "name" in mutation

    def test_delete_mutation_default_fields(self):
        """Test delete mutation with default fields (only id)"""
        mutation = MutationBuilder.build_delete_mutation("Product")

        assert "mutation DeleteProduct" in mutation
        assert "$input: DeleteProductInput!" in mutation
        assert "deleteProduct(input: $input)" in mutation
        assert "id" in mutation

    def test_delete_mutation_single_field(self):
        """Test delete mutation with single return field"""
        mutation = MutationBuilder.build_delete_mutation("Comment", return_fields=["id"])

        assert "mutation DeleteComment" in mutation
        assert "id" in mutation

    def test_delete_mutation_different_models(self):
        """Test delete mutation with different model names"""
        models = ["User", "Product", "Order", "Comment"]

        for model in models:
            mutation = MutationBuilder.build_delete_mutation(model)
            assert f"mutation Delete{model}" in mutation
            assert f"$input: Delete{model}Input!" in mutation
            assert f"delete{model}(input: $input)" in mutation


class TestBuildCreateVariables:
    """Test MutationBuilder.build_create_variables() method"""

    def test_create_variables_simple(self):
        """Test create variables with simple input"""
        input_data = {"name": "John", "email": "john@example.com"}
        variables = MutationBuilder.build_create_variables(input_data)

        assert variables == {"input": {"name": "John", "email": "john@example.com"}}

    def test_create_variables_complex(self):
        """Test create variables with complex input"""
        input_data = {
            "name": "Alice",
            "email": "alice@example.com",
            "age": 30,
            "isActive": True,
            "settings": {"theme": "dark", "notifications": True},
        }
        variables = MutationBuilder.build_create_variables(input_data)

        assert variables == {"input": input_data}
        assert variables["input"]["settings"]["theme"] == "dark"

    def test_create_variables_empty(self):
        """Test create variables with empty input"""
        input_data = {}
        variables = MutationBuilder.build_create_variables(input_data)

        assert variables == {"input": {}}

    def test_create_variables_with_nested_objects(self):
        """Test create variables with nested objects"""
        input_data = {
            "title": "Test Post",
            "author": {"name": "John", "email": "john@test.com"},
            "tags": ["python", "testing"],
        }
        variables = MutationBuilder.build_create_variables(input_data)

        assert variables["input"]["author"]["name"] == "John"
        assert "python" in variables["input"]["tags"]


class TestBuildUpdateVariables:
    """Test MutationBuilder.build_update_variables() method"""

    def test_update_variables_basic(self):
        """Test update variables with basic updates"""
        variables = MutationBuilder.build_update_variables("user-123", {"name": "Jane", "email": "jane@example.com"})

        assert variables == {"input": {"id": "user-123", "name": "Jane", "email": "jane@example.com"}}

    def test_update_variables_single_field(self):
        """Test update variables with single field update"""
        variables = MutationBuilder.build_update_variables("product-456", {"price": 99.99})

        assert variables == {"input": {"id": "product-456", "price": 99.99}}

    def test_update_variables_empty_updates(self):
        """Test update variables with empty updates (only ID)"""
        variables = MutationBuilder.build_update_variables("order-789", {})

        assert variables == {"input": {"id": "order-789"}}

    def test_update_variables_complex_updates(self):
        """Test update variables with complex updates"""
        updates = {"status": "published", "updatedAt": "2024-01-01T00:00:00Z", "tags": ["featured", "trending"]}
        variables = MutationBuilder.build_update_variables("post-111", updates)

        assert variables["input"]["id"] == "post-111"
        assert variables["input"]["status"] == "published"
        assert "featured" in variables["input"]["tags"]

    def test_update_variables_id_not_overwritten(self):
        """Test that ID in updates doesn't overwrite record_id parameter"""
        # Even if updates contains an 'id', the record_id parameter should take precedence
        updates = {"id": "wrong-id", "name": "Test"}
        variables = MutationBuilder.build_update_variables("correct-id", updates)

        # The record_id parameter is added last, so it should overwrite
        # Actually, the current implementation adds id first, then spreads updates
        # So updates would overwrite. Let's test the actual behavior
        assert variables["input"]["id"] == "wrong-id"  # This is what actually happens
        assert variables["input"]["name"] == "Test"

    def test_update_variables_preserves_data_types(self):
        """Test that update variables preserve data types"""
        updates = {"name": "Test", "age": 25, "isActive": True, "score": 95.5, "tags": ["a", "b"]}
        variables = MutationBuilder.build_update_variables("user-999", updates)

        assert isinstance(variables["input"]["age"], int)
        assert isinstance(variables["input"]["isActive"], bool)
        assert isinstance(variables["input"]["score"], float)
        assert isinstance(variables["input"]["tags"], list)


class TestBuildDeleteVariables:
    """Test MutationBuilder.build_delete_variables() method"""

    def test_delete_variables_basic(self):
        """Test delete variables with ID"""
        variables = MutationBuilder.build_delete_variables("user-123")

        assert variables == {"input": {"id": "user-123"}}

    def test_delete_variables_different_ids(self):
        """Test delete variables with different ID formats"""
        ids = ["user-123", "product-abc", "order-xyz-789", "comment-111-222-333"]

        for record_id in ids:
            variables = MutationBuilder.build_delete_variables(record_id)
            assert variables == {"input": {"id": record_id}}

    def test_delete_variables_numeric_id(self):
        """Test delete variables with numeric ID (as string)"""
        variables = MutationBuilder.build_delete_variables("12345")

        assert variables == {"input": {"id": "12345"}}


class TestMutationBuilderIntegration:
    """Integration tests for MutationBuilder"""

    def test_create_mutation_builds_valid_graphql(self):
        """Test that create mutation builds syntactically valid GraphQL"""
        mutation = MutationBuilder.build_create_mutation("User", return_fields=["id", "name"])

        # Should have matching braces
        assert mutation.count("{") == mutation.count("}")
        # Should have mutation keyword
        assert mutation.startswith("mutation ")

    def test_update_mutation_builds_valid_graphql(self):
        """Test that update mutation builds syntactically valid GraphQL"""
        mutation = MutationBuilder.build_update_mutation("Product", return_fields=["id", "title"])

        # Should have matching braces
        assert mutation.count("{") == mutation.count("}")
        # Should have mutation keyword
        assert mutation.startswith("mutation ")

    def test_delete_mutation_builds_valid_graphql(self):
        """Test that delete mutation builds syntactically valid GraphQL"""
        mutation = MutationBuilder.build_delete_mutation("Order", return_fields=["id"])

        # Should have matching braces
        assert mutation.count("{") == mutation.count("}")
        # Should have mutation keyword
        assert mutation.startswith("mutation ")

    def test_mutations_dont_contain_query_keyword(self):
        """Test that mutations don't contain query keywords"""
        create = MutationBuilder.build_create_mutation("User")
        update = MutationBuilder.build_update_mutation("User")
        delete = MutationBuilder.build_delete_mutation("User")

        for mutation in [create, update, delete]:
            # Should not have 'query' keyword (except in comments)
            assert mutation.startswith("mutation ")
            assert "mutation" in mutation.lower()

    def test_field_indentation_consistent(self):
        """Test that fields are properly indented in mutations"""
        mutation = MutationBuilder.build_create_mutation("User", return_fields=["id", "name", "email"])

        # Fields should be indented (have leading spaces)
        lines = mutation.split("\n")
        field_lines = [line for line in lines if line.strip() in ["id", "name", "email"]]

        for line in field_lines:
            # Fields should be indented
            assert line.startswith("        ")  # 8 spaces

    def test_input_type_names_match_model(self):
        """Test that input type names match the model name"""
        models = ["User", "Product", "Order"]

        for model in models:
            create = MutationBuilder.build_create_mutation(model)
            update = MutationBuilder.build_update_mutation(model)
            delete = MutationBuilder.build_delete_mutation(model)

            assert f"Create{model}Input" in create
            assert f"Update{model}Input" in update
            assert f"Delete{model}Input" in delete

    def test_variables_structure_matches_mutations(self):
        """Test that variable builders create structure expected by mutations"""
        # Create
        create_vars = MutationBuilder.build_create_variables({"name": "Test"})
        assert "input" in create_vars
        assert isinstance(create_vars["input"], dict)

        # Update
        update_vars = MutationBuilder.build_update_variables("id-123", {"name": "Test"})
        assert "input" in update_vars
        assert "id" in update_vars["input"]

        # Delete
        delete_vars = MutationBuilder.build_delete_variables("id-456")
        assert "input" in delete_vars
        assert delete_vars["input"]["id"] == "id-456"

    def test_all_mutations_use_input_variable(self):
        """Test that all mutations use $input variable"""
        create = MutationBuilder.build_create_mutation("User")
        update = MutationBuilder.build_update_mutation("User")
        delete = MutationBuilder.build_delete_mutation("User")

        for mutation in [create, update, delete]:
            assert "$input:" in mutation
            assert "(input: $input)" in mutation


class TestMutationBuilderEdgeCases:
    """Test edge cases for MutationBuilder"""

    def test_model_name_with_multiple_words(self):
        """Test mutations with multi-word model names"""
        mutation = MutationBuilder.build_create_mutation("BlogPost")

        assert "mutation CreateBlogPost" in mutation
        assert "CreateBlogPostInput" in mutation
        assert "createBlogPost(input: $input)" in mutation

    def test_empty_return_fields_list(self):
        """Test mutation with empty return fields list"""
        # Even with empty list, should work (though not useful)
        mutation = MutationBuilder.build_create_mutation("User", return_fields=[])

        assert "mutation CreateUser" in mutation
        # Should still have the mutation structure even if no fields returned

    def test_return_fields_with_special_characters(self):
        """Test mutation with field names containing underscores"""
        fields = ["id", "created_at", "updated_at", "user_id"]
        mutation = MutationBuilder.build_create_mutation("Post", return_fields=fields)

        for field in fields:
            assert field in mutation

    def test_large_number_of_return_fields(self):
        """Test mutation with many return fields"""
        fields = [f"field{i}" for i in range(50)]
        mutation = MutationBuilder.build_create_mutation("TestModel", return_fields=fields)

        for field in fields:
            assert field in mutation

    def test_variables_with_none_values(self):
        """Test create variables with None values"""
        input_data = {"name": "Test", "description": None, "age": 0}
        variables = MutationBuilder.build_create_variables(input_data)

        assert variables["input"]["description"] is None
        assert variables["input"]["age"] == 0  # 0 should be preserved, not treated as falsy



import asyncio
import logging
import sys
from typing import Dict, Any, Optional

import aiohttp
import requests
import inflect

from amplify_excel_migrator.graphql import QueryBuilder, MutationBuilder
from amplify_excel_migrator.auth import AuthenticationProvider

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


class AuthenticationError(Exception):
    """Raised when authentication is required but not completed"""

    pass


class GraphQLError(Exception):
    """Raised when GraphQL query returns errors"""

    pass


class AmplifyClient:
    """
    Client for Amplify GraphQL API operations.
    """

    def __init__(self, api_endpoint: str, auth_provider: Optional[AuthenticationProvider] = None):
        """
        Initialize the client

        Args:
            api_endpoint: Amplify GraphQL endpoint
            auth_provider: Authentication provider instance
        """
        self.api_endpoint = api_endpoint
        self.auth_provider = auth_provider

        self.batch_size = 20
        self.records_cache = {}

    def _request(self, query: str, variables: Dict = None, context: str = None) -> Any | None:
        """
        Make a GraphQL request using the ID token

        Args:
            query: GraphQL query or mutation
            variables: Variables for the query
            context: Optional context string to include in error messages (e.g., row identifier)

        Returns:
            Response data
        """
        if not self.auth_provider or not self.auth_provider.is_authenticated():
            raise AuthenticationError("Not authenticated. Call authenticate() on the auth provider first.")

        id_token = self.auth_provider.get_id_token()
        headers = {"Authorization": id_token, "Content-Type": "application/json"}

        payload = {"query": query, "variables": variables or {}}

        context_msg = f" [{context}]" if context else ""

        try:
            response = requests.post(self.api_endpoint, headers=headers, json=payload)

            if response.status_code == 200:
                result = response.json()

                if "errors" in result:
                    raise GraphQLError(f"GraphQL errors{context_msg}: {result['errors']}")

                return result
            else:
                logger.error(f"HTTP Error {response.status_code}{context_msg}: {response.text}")
                return None

        except requests.exceptions.ConnectionError as e:
            logger.error(
                f"Connection error{context_msg}: Unable to connect to API endpoint. Check your internet connection or the API endpoint URL."
            )
            sys.exit(1)

        except requests.exceptions.Timeout as e:
            logger.error(f"Request timeout{context_msg}: {e}")
            return None

        except requests.exceptions.HTTPError as e:
            logger.error(f"HTTP error{context_msg}: {e}")
            return None

        except GraphQLError as e:
            logger.error(str(e))
            return None

        except requests.exceptions.RequestException as e:
            logger.error(f"Request error{context_msg}: {e}")
            return None

    async def _request_async(
        self, session: aiohttp.ClientSession, query: str, variables: Dict = None, context: str = None
    ) -> Any | None:
        """
        Async version of _request for parallel GraphQL requests

        Args:
            session: aiohttp ClientSession
            query: GraphQL query or mutation
            variables: Variables for the query
            context: Optional context string to include in error messages (e.g., row identifier)

        Returns:
            Response data
        """
        if not self.auth_provider or not self.auth_provider.is_authenticated():
            raise AuthenticationError("Not authenticated. Call authenticate() on the auth provider first.")

        id_token = self.auth_provider.get_id_token()
        headers = {"Authorization": id_token, "Content-Type": "application/json"}

        payload = {"query": query, "variables": variables or {}}

        context_msg = f" [{context}]" if context else ""

        try:
            async with session.post(self.api_endpoint, headers=headers, json=payload) as response:
                if response.status == 200:
                    result = await response.json()

                    if "errors" in result:
                        raise GraphQLError(f"GraphQL errors{context_msg}: {result['errors']}")

                    return result
                else:
                    text = await response.text()
                    error_msg = f"HTTP Error {response.status}{context_msg}: {text}"
                    logger.error(error_msg)
                    raise aiohttp.ClientError(error_msg)

        except aiohttp.ServerTimeoutError as e:
            error_msg = f"Request timeout{context_msg}: {e}"
            logger.error(error_msg)
            raise aiohttp.ServerTimeoutError(error_msg)

        except aiohttp.ClientConnectionError as e:
            error_msg = f"Connection error{context_msg}: Unable to connect to API endpoint. {e}"
            logger.error(error_msg)
            raise aiohttp.ClientConnectionError(error_msg)

        except aiohttp.ClientResponseError as e:
            error_msg = f"HTTP response error{context_msg}: {e}"
            logger.error(error_msg)
            raise aiohttp.ClientResponseError(
                request_info=e.request_info, history=e.history, status=e.status, message=error_msg
            )

        except GraphQLError as e:
            logger.error(str(e))
            raise

        except aiohttp.ClientError as e:
            error_msg = f"Client error{context_msg}: {e}"
            logger.error(error_msg)
            raise aiohttp.ClientError(error_msg)

    async def create_record_async(
        self, session: aiohttp.ClientSession, data: Dict, model_name: str, primary_field: str
    ) -> Dict | None:
        mutation = MutationBuilder.build_create_mutation(model_name, return_fields=["id", primary_field])
        variables = MutationBuilder.build_create_variables(data)

        context = f"{model_name}: {primary_field}={data.get(primary_field)}"
        result = await self._request_async(session, mutation, variables, context)

        if result and "data" in result:
            created = result["data"].get(f"create{model_name}")
            if created:
                logger.info(f'Created {model_name} with {primary_field}="{data[primary_field]}" (ID: {created["id"]})')
            return created
        else:
            logger.error(f'Failed to create {model_name} with {primary_field}="{data[primary_field]}"')

        return None

    async def check_record_exists_async(
        self,
        session: aiohttp.ClientSession,
        model_name: str,
        primary_field: str,
        value: str,
        is_secondary_index: bool,
        record: Dict,
        field_type: str = "String",
    ) -> Dict | None:
        context = f"{model_name}: {primary_field}={value}"

        if is_secondary_index:
            query = QueryBuilder.build_secondary_index_query(
                model_name, primary_field, fields=["id"], field_type=field_type, with_pagination=False
            )
            variables = {primary_field: value}
            query_name = f"list{model_name}By{primary_field[0].upper() + primary_field[1:]}"

            result = await self._request_async(session, query, variables, context)
            if result and "data" in result:
                items = result["data"].get(query_name, {}).get("items", [])
                if len(items) > 0:
                    logger.warning(f'Record with {primary_field}="{value}" already exists in {model_name}')
                    return None
        else:
            query_name = self._get_list_query_name(model_name)
            query = QueryBuilder.build_list_query_with_filter(model_name, fields=["id"], with_pagination=False)
            filter_input = QueryBuilder.build_filter_equals(primary_field, value)
            variables = {"filter": filter_input}

            result = await self._request_async(session, query, variables, context)
            if result and "data" in result:
                items = result["data"].get(query_name, {}).get("items", [])
                if len(items) > 0:
                    logger.error(f'Record with {primary_field}="{value}" already exists in {model_name}')
                    return None

        return record

    async def upload_batch_async(
        self, batch: list, model_name: str, primary_field: str, is_secondary_index: bool, field_type: str = "String"
    ) -> tuple[int, int, list[Dict]]:
        async with aiohttp.ClientSession() as session:
            duplicate_checks = [
                self.check_record_exists_async(
                    session, model_name, primary_field, record[primary_field], is_secondary_index, record, field_type
                )
                for record in batch
            ]
            check_results = await asyncio.gather(*duplicate_checks, return_exceptions=True)

            filtered_batch = []
            failed_records = []

            for i, result in enumerate(check_results):
                if isinstance(result, Exception):
                    error_msg = str(result)
                    failed_records.append(
                        {
                            "primary_field": primary_field,
                            "primary_field_value": batch[i].get(primary_field, "Unknown"),
                            "error": f"Duplicate check error: {error_msg}",
                        }
                    )
                    logger.error(f"Error checking duplicate: {result}")
                elif result is not None:
                    filtered_batch.append(result)

            if not filtered_batch:
                return 0, len(batch), failed_records

            create_tasks = [
                self.create_record_async(session, record, model_name, primary_field) for record in filtered_batch
            ]
            results = await asyncio.gather(*create_tasks, return_exceptions=True)

            for i, result in enumerate(results):
                if isinstance(result, Exception):
                    error_msg = str(result)
                    failed_records.append(
                        {
                            "primary_field": primary_field,
                            "primary_field_value": filtered_batch[i].get(primary_field, "Unknown"),
                            "error": error_msg,
                        }
                    )
                elif not result:
                    failed_records.append(
                        {
                            "primary_field": primary_field,
                            "primary_field_value": filtered_batch[i].get(primary_field, "Unknown"),
                            "error": "Creation failed - no response",
                        }
                    )

            success_count = sum(1 for r in results if r and not isinstance(r, Exception))
            error_count = len(batch) - success_count

            return success_count, error_count, failed_records

    def get_model_structure(self, model_type: str) -> Dict:
        query = f"""
        query GetModelType {{
          __type(name: "{model_type}") {{
            name
            kind
            description
            fields {{
              name
              type {{
                name
                kind
                ofType {{
                  name
                  kind
                  ofType {{
                    name
                    kind
                  }}
                }}
              }}
              description
            }}
          }}
        }}
        """

        response = self._request(query)
        if response and "data" in response and "__type" in response["data"]:
            return response["data"]["__type"]

        return {}

    def get_primary_field_name(self, model_name: str, parsed_model_structure: Dict[str, Any]) -> tuple[str, bool, str]:
        """
        Returns: (field_name, is_secondary_index, field_type)
        """
        secondary_index = self._get_secondary_index(model_name)
        if secondary_index:
            field_type = "String"
            for field in parsed_model_structure["fields"]:
                if field["name"] == secondary_index:
                    field_type = field["type"]
                    break
            return secondary_index, True, field_type

        for field in parsed_model_structure["fields"]:
            if field["is_required"] and field["is_scalar"] and field["name"] != "id":
                return field["name"], False, field["type"]

        logger.error("No suitable primary field found (required scalar field other than id)")
        return "", False, "String"

    def _get_secondary_index(self, model_name: str) -> str:
        query_structure = self.get_model_structure("Query")
        if not query_structure:
            logger.error("Query type not found in schema")
            return ""

        query_fields = query_structure["fields"]

        pattern = f"{model_name}By"

        for query in query_fields:
            query_name = query["name"]
            if pattern in query_name:
                pattern_index = query_name.index(pattern)
                field_name = query_name[pattern_index + len(pattern) :]
                return field_name[0].lower() + field_name[1:] if field_name else ""

        return ""

    def _get_list_query_name(self, model_name: str) -> str | None:
        query_structure = self.get_model_structure("Query")
        if not query_structure:
            logger.error("Query type not found in schema")
            return f"list{model_name}s"

        query_fields = query_structure["fields"]
        p = inflect.engine()

        candidates = [f"list{model_name}"]
        capitals = [i for i, c in enumerate(model_name) if c.isupper()]

        if len(capitals) > 1:
            last_word_start = capitals[-1]
            prefix = model_name[:last_word_start]
            last_word = model_name[last_word_start:]

            last_word_plural = str(p.plural(last_word.lower()))  # type: ignore[arg-type]
            last_word_plural_cap = last_word_plural[0].upper() + last_word_plural[1:] if last_word_plural else ""

            pascal_plural = f"{prefix}{last_word_plural_cap}"
            candidates.append(f"list{pascal_plural}")

        full_plural = str(p.plural(model_name.lower()))  # type: ignore[arg-type]
        full_plural_cap = full_plural[0].upper() + full_plural[1:] if full_plural else ""
        candidates.append(f"list{full_plural_cap}")

        for query in query_fields:
            query_name = query["name"]
            if query_name in candidates and "By" not in query_name:
                return query_name

        logger.error(f"No list query found for model {model_name}, tried: {candidates}")
        return None

    def upload(
        self, records: list, model_name: str, parsed_model_structure: Dict[str, Any]
    ) -> tuple[int, int, list[Dict]]:
        logger.info("Uploading to Amplify backend...")

        success_count = 0
        error_count = 0
        all_failed_records = []
        num_of_batches = (len(records) + self.batch_size - 1) // self.batch_size

        primary_field, is_secondary_index, field_type = self.get_primary_field_name(model_name, parsed_model_structure)
        if not primary_field:
            logger.error(f"Aborting upload for model {model_name}")
            return 0, len(records), []

        for i in range(0, len(records), self.batch_size):
            batch = records[i : i + self.batch_size]
            logger.info(f"Uploading batch {i // self.batch_size + 1} / {num_of_batches} ({len(batch)} items)...")

            batch_success, batch_error, batch_failed_records = asyncio.run(
                self.upload_batch_async(batch, model_name, primary_field, is_secondary_index, field_type)
            )
            success_count += batch_success
            error_count += batch_error
            all_failed_records.extend(batch_failed_records)

            logger.info(
                f"Processed batch {i // self.batch_size + 1} of model {model_name}: {success_count} success, {error_count} errors"
            )

        return success_count, error_count, all_failed_records

    def list_records_by_secondary_index(
        self, model_name: str, secondary_index: str, value: str = None, fields: list = None, field_type: str = "String"
    ) -> Dict | None:
        if fields is None:
            fields = ["id", secondary_index]

        fields_str = "\n".join(fields)
        all_items = []
        next_token = None

        if not value:
            query_name = self._get_list_query_name(model_name)

            while True:
                query = f"""
                query List{model_name}s($limit: Int, $nextToken: String) {{
                  {query_name}(limit: $limit, nextToken: $nextToken) {{
                    items {{
                        {fields_str}
                    }}
                    nextToken
                  }}
                }}
                """
                variables = {"limit": 1000, "nextToken": next_token}
                result = self._request(query, variables)

                if result and "data" in result:
                    data = result["data"].get(query_name, {})
                    items = data.get("items", [])
                    all_items.extend(items)
                    next_token = data.get("nextToken")

                    if not next_token:
                        break
                else:
                    break
        else:
            query_name = f"list{model_name}By{secondary_index[0].upper() + secondary_index[1:]}"

            while True:
                query = f"""
                query {query_name}(${secondary_index}: {field_type}!, $limit: Int, $nextToken: String) {{
                  {query_name}({secondary_index}: ${secondary_index}, limit: $limit, nextToken: $nextToken) {{
                    items {{
                        {fields_str}
                    }}
                    nextToken
                  }}
                }}
                """
                variables = {secondary_index: value, "limit": 1000, "nextToken": next_token}
                result = self._request(query, variables)

                if result and "data" in result:
                    data = result["data"].get(query_name, {})
                    items = data.get("items", [])
                    all_items.extend(items)
                    next_token = data.get("nextToken")

                    if not next_token:
                        break
                else:
                    break

        return all_items if all_items else None

    def list_records_by_field(
        self, model_name: str, field_name: str, value: str = None, fields: list = None
    ) -> Dict | None:
        if fields is None:
            fields = ["id", field_name]

        fields_str = "\n".join(fields)
        all_items = []
        next_token = None

        query_name = self._get_list_query_name(model_name)

        if not value:
            while True:
                query = f"""
                query List{model_name}s($limit: Int, $nextToken: String) {{
                  {query_name}(limit: $limit, nextToken: $nextToken) {{
                    items {{
                        {fields_str}
                    }}
                    nextToken
                  }}
                }}
                """
                variables = {"limit": 1000, "nextToken": next_token}
                result = self._request(query, variables)

                if result and "data" in result:
                    data = result["data"].get(query_name, {})
                    items = data.get("items", [])
                    all_items.extend(items)
                    next_token = data.get("nextToken")

                    if not next_token:
                        break
                else:
                    break
        else:
            while True:
                query = f"""
                query List{model_name}s($filter: Model{model_name}FilterInput, $limit: Int, $nextToken: String) {{
                  {query_name}(filter: $filter, limit: $limit, nextToken: $nextToken) {{
                    items {{
                        {fields_str}
                    }}
                    nextToken
                  }}
                }}
                """
                filter_input = {field_name: {"eq": value}}
                variables = {"filter": filter_input, "limit": 1000, "nextToken": next_token}
                result = self._request(query, variables)

                if result and "data" in result:
                    data = result["data"].get(query_name, {})
                    items = data.get("items", [])
                    all_items.extend(items)
                    next_token = data.get("nextToken")

                    if not next_token:
                        break
                else:
                    break

        return all_items if all_items else None

    def get_record_by_id(self, model_name: str, record_id: str, fields: list = None) -> Dict | None:
        if fields is None:
            fields = ["id"]

        query = QueryBuilder.build_get_by_id_query(model_name, fields=fields)
        query_name = f"get{model_name}"

        result = self._request(query, {"id": record_id})

        if result and "data" in result:
            return result["data"].get(query_name)

        return None

    def get_records(
        self,
        model_name: str,
        primary_field: str = None,
        is_secondary_index: bool = None,
        fields: list = None,
    ) -> list | None:
        if model_name in self.records_cache:
            return self.records_cache[model_name]

        if not primary_field:
            return None
        if is_secondary_index:
            records = self.list_records_by_secondary_index(model_name, primary_field, fields=fields)
        else:
            records = self.list_records_by_field(model_name, primary_field, fields=fields)

        if records:
            self.records_cache[model_name] = records
            logger.debug(f"💾 Cached {len(records)} records for {model_name}")
        return records

    def get_record(
        self,
        model_name: str,
        parsed_model_structure: Dict[str, Any] = None,
        value: str = None,
        record_id: str = None,
        primary_field: str = None,
        is_secondary_index: bool = None,
        fields: list = None,
    ) -> Dict | None:
        if record_id:
            return self.get_record_by_id(model_name, record_id)

        if not primary_field:
            if not parsed_model_structure:
                logger.error("Parsed model structure required if primary_field not provided")
                return None
            primary_field, is_secondary_index, _ = self.get_primary_field_name(model_name, parsed_model_structure)
        records = self.get_records(model_name, primary_field, is_secondary_index, fields)
        if not records:
            return None
        return next((record for record in records if record.get(primary_field) == value), None)

    def build_foreign_key_lookups(self, df, parsed_model_structure: Dict[str, Any]) -> Dict[str, Dict[str, str]]:
        """
        Build a cache of foreign key lookups for all ID fields in the DataFrame.

        This pre-fetches all related records to avoid N+1 query problems during row processing.

        Args:
            df: pandas DataFrame containing the data to be processed
            parsed_model_structure: Parsed model structure containing field information

        Returns:
            Dictionary mapping model names to lookup dictionaries and primary fields
        """

        fk_lookup_cache = {}

        for field in parsed_model_structure["fields"]:
            if not field["is_id"]:
                continue

            field_name = field["name"][:-2]

            if field_name not in df.columns:
                continue

            if "related_model" in field:
                related_model = field["related_model"]
            else:
                related_model = field_name[0].upper() + field_name[1:]

            if related_model in fk_lookup_cache:
                continue

            try:
                primary_field, is_secondary_index, _ = self.get_primary_field_name(
                    related_model, parsed_model_structure
                )
                records = self.get_records(related_model, primary_field, is_secondary_index)

                if records:
                    lookup = {
                        str(record.get(primary_field)): record.get("id")
                        for record in records
                        if record.get(primary_field)
                    }
                    fk_lookup_cache[related_model] = {"lookup": lookup, "primary_field": primary_field}
                    logger.debug(f"  📦 Cached {len(lookup)} {related_model} records")
            except Exception as e:
                logger.warning(f"  ⚠️  Could not pre-fetch {related_model}: {e}")

        return fk_lookup_cache



"""Tests for CognitoAuthProvider class"""

import pytest
from unittest.mock import Mock, patch, MagicMock
from pycognito.exceptions import ForceChangePasswordException
from pycognito import MFAChallengeException
from botocore.exceptions import NoCredentialsError, ProfileNotFound, NoRegionError, ClientError

from amplify_excel_migrator.auth import CognitoAuthProvider


@pytest.fixture
def auth_params():
    """Standard authentication parameters"""
    return {
        "user_pool_id": "us-east-1_testpool",
        "client_id": "test-client-id",
        "region": "us-east-1",
        "admin_group_name": "ADMINS",
    }


@pytest.fixture
def cognito_provider(auth_params):
    """Create CognitoAuthProvider instance"""
    return CognitoAuthProvider(**auth_params)


@pytest.fixture
def mock_id_token():
    """Mock ID token with ADMINS group"""
    return "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwiY29nbml0bzpncm91cHMiOlsiQURNSU5TIl19.test"


class TestCognitoAuthProviderInitialization:
    """Test CognitoAuthProvider initialization"""

    def test_initialization(self, cognito_provider, auth_params):
        """Test basic initialization"""
        assert cognito_provider.user_pool_id == auth_params["user_pool_id"]
        assert cognito_provider.client_id == auth_params["client_id"]
        assert cognito_provider.region == auth_params["region"]
        assert cognito_provider.admin_group_name == auth_params["admin_group_name"]
        assert cognito_provider.cognito_client is None
        assert cognito_provider.boto_cognito_admin_client is None
        assert cognito_provider._id_token is None
        assert cognito_provider._mfa_tokens is None

    def test_custom_admin_group(self, auth_params):
        """Test initialization with custom admin group"""
        auth_params["admin_group_name"] = "CUSTOM_ADMINS"
        provider = CognitoAuthProvider(**auth_params)
        assert provider.admin_group_name == "CUSTOM_ADMINS"


class TestStandardAuthentication:
    """Test standard Cognito authentication"""

    @patch("amplify_excel_migrator.auth.cognito_auth.Cognito")
    @patch("amplify_excel_migrator.auth.cognito_auth.jwt.decode")
    def test_successful_authentication(self, mock_jwt_decode, mock_cognito_class, cognito_provider, mock_id_token):
        """Test successful authentication"""
        mock_cognito = Mock()
        mock_cognito.id_token = mock_id_token
        mock_cognito_class.return_value = mock_cognito
        mock_jwt_decode.return_value = {"cognito:groups": ["ADMINS"]}

        result = cognito_provider.authenticate("test@example.com", "password123")

        assert result is True
        assert cognito_provider._id_token == mock_id_token
        mock_cognito.authenticate.assert_called_once_with(password="password123")

    @patch("amplify_excel_migrator.auth.cognito_auth.Cognito")
    @patch("amplify_excel_migrator.auth.cognito_auth.jwt.decode")
    def test_authentication_not_in_admin_group(self, mock_jwt_decode, mock_cognito_class, cognito_provider):
        """Test authentication fails when user not in admin group"""
        mock_cognito = Mock()
        mock_cognito.id_token = "test_token"
        mock_cognito_class.return_value = mock_cognito
        mock_jwt_decode.return_value = {"cognito:groups": ["USERS"]}

        result = cognito_provider.authenticate("test@example.com", "password123")
        assert result is False

    @patch("amplify_excel_migrator.auth.cognito_auth.Cognito")
    @patch("amplify_excel_migrator.auth.cognito_auth.input")
    def test_authentication_with_mfa_challenge(self, mock_input, mock_cognito_class, cognito_provider, mock_id_token):
        """Test authentication with MFA challenge"""
        mock_cognito = Mock()
        mock_cognito_class.return_value = mock_cognito

        mfa_tokens = {"ChallengeName": "SMS_MFA", "Session": "test-session"}
        mfa_exception = MFAChallengeException("MFA required", mfa_tokens)

        mock_cognito.authenticate.side_effect = [mfa_exception, None]
        mock_cognito.id_token = mock_id_token
        mock_input.return_value = "123456"

        with patch.object(cognito_provider, "_complete_mfa_challenge", return_value=True):
            with patch.object(cognito_provider, "_check_user_in_admins_group"):
                result = cognito_provider.authenticate("test@example.com", "password123")

        assert result is True

    @patch("amplify_excel_migrator.auth.cognito_auth.Cognito")
    @patch("amplify_excel_migrator.auth.cognito_auth.input")
    def test_authentication_with_password_change(self, mock_input, mock_cognito_class, cognito_provider, mock_id_token):
        """Test authentication with force password change"""
        mock_cognito = Mock()
        mock_cognito_class.return_value = mock_cognito

        mock_cognito.authenticate.side_effect = [ForceChangePasswordException(), None]
        mock_cognito.id_token = mock_id_token
        mock_input.side_effect = ["newpassword123", "newpassword123"]

        with patch.object(cognito_provider, "_check_user_in_admins_group"):
            result = cognito_provider.authenticate("test@example.com", "oldpassword")

        assert result is True
        mock_cognito.new_password_challenge.assert_called_once()

    @patch("amplify_excel_migrator.auth.cognito_auth.Cognito")
    @patch("amplify_excel_migrator.auth.cognito_auth.input")
    def test_authentication_password_mismatch(self, mock_input, mock_cognito_class, cognito_provider):
        """Test authentication fails when new passwords don't match"""
        mock_cognito = Mock()
        mock_cognito_class.return_value = mock_cognito
        mock_cognito.authenticate.side_effect = ForceChangePasswordException()
        mock_input.side_effect = ["newpassword", "differentpassword"]

        result = cognito_provider.authenticate("test@example.com", "oldpassword")

        assert result is False

    @patch("amplify_excel_migrator.auth.cognito_auth.Cognito")
    def test_authentication_general_error(self, mock_cognito_class, cognito_provider):
        """Test authentication handles general errors"""
        mock_cognito = Mock()
        mock_cognito_class.return_value = mock_cognito
        mock_cognito.authenticate.side_effect = Exception("Network error")

        result = cognito_provider.authenticate("test@example.com", "password123")

        assert result is False


class TestAdminAuthentication:
    """Test admin authentication"""

    @patch("amplify_excel_migrator.auth.cognito_auth.boto3.client")
    @patch("amplify_excel_migrator.auth.cognito_auth.jwt.decode")
    def test_successful_admin_authentication(self, mock_jwt_decode, mock_boto_client, cognito_provider, mock_id_token):
        """Test successful admin authentication"""
        mock_cognito_admin = Mock()
        mock_boto_client.return_value = mock_cognito_admin
        mock_jwt_decode.return_value = {"cognito:groups": ["ADMINS"]}

        mock_cognito_admin.admin_initiate_auth.return_value = {
            "AuthenticationResult": {"IdToken": mock_id_token, "AccessToken": "access_token"}
        }

        result = cognito_provider.authenticate_admin("test@example.com", "password123")

        assert result is True
        assert cognito_provider._id_token == mock_id_token
        mock_cognito_admin.admin_initiate_auth.assert_called_once()

    @patch("amplify_excel_migrator.auth.cognito_auth.boto3.Session")
    def test_admin_authentication_with_profile(self, mock_session_class, cognito_provider, mock_id_token):
        """Test admin authentication with AWS profile"""
        mock_session = Mock()
        mock_session_class.return_value = mock_session
        mock_cognito_admin = Mock()
        mock_session.client.return_value = mock_cognito_admin

        mock_cognito_admin.admin_initiate_auth.return_value = {"AuthenticationResult": {"IdToken": mock_id_token}}

        with patch("amplify_excel_migrator.auth.cognito_auth.jwt.decode") as mock_jwt:
            mock_jwt.return_value = {"cognito:groups": ["ADMINS"]}
            result = cognito_provider.authenticate_admin("test@example.com", "password", aws_profile="my-profile")

        assert result is True
        mock_session_class.assert_called_once_with(profile_name="my-profile")

    @patch("amplify_excel_migrator.auth.cognito_auth.boto3.client")
    def test_admin_authentication_no_credentials(self, mock_boto_client, cognito_provider):
        """Test admin authentication fails with no AWS credentials"""
        mock_boto_client.side_effect = NoCredentialsError()

        result = cognito_provider.authenticate_admin("test@example.com", "password123")
        assert result is False

    @patch("amplify_excel_migrator.auth.cognito_auth.boto3.Session")
    def test_admin_authentication_profile_not_found(self, mock_session_class, cognito_provider):
        """Test admin authentication fails with invalid profile"""
        mock_session_class.side_effect = ProfileNotFound(profile="invalid")

        result = cognito_provider.authenticate_admin("test@example.com", "password", aws_profile="invalid")
        assert result is False

    @patch("amplify_excel_migrator.auth.cognito_auth.boto3.client")
    def test_admin_authentication_no_region(self, mock_boto_client, cognito_provider):
        """Test admin authentication fails with no region"""
        mock_boto_client.side_effect = NoRegionError()

        result = cognito_provider.authenticate_admin("test@example.com", "password123")
        assert result is False

    @patch("amplify_excel_migrator.auth.cognito_auth.boto3.client")
    def test_admin_authentication_client_error(self, mock_boto_client, cognito_provider):
        """Test admin authentication handles AWS client errors"""
        error_response = {"Error": {"Code": "AccessDenied", "Message": "Access denied"}}
        mock_boto_client.side_effect = ClientError(error_response, "admin_initiate_auth")

        result = cognito_provider.authenticate_admin("test@example.com", "password123")
        assert result is False

    @patch("amplify_excel_migrator.auth.cognito_auth.boto3.client")
    def test_admin_authentication_no_result(self, mock_boto_client, cognito_provider):
        """Test admin authentication fails when no AuthenticationResult"""
        mock_cognito_admin = Mock()
        mock_boto_client.return_value = mock_cognito_admin
        mock_cognito_admin.admin_initiate_auth.return_value = {"ChallengeName": "MFA_SETUP"}

        with patch.object(cognito_provider, "_check_for_mfa_challenges", return_value=False):
            result = cognito_provider.authenticate_admin("test@example.com", "password123")

        assert result is False


class TestMFAChallenge:
    """Test MFA challenge handling"""

    def test_complete_sms_mfa_challenge(self, cognito_provider):
        """Test completing SMS MFA challenge"""
        mock_cognito = Mock()
        cognito_provider.cognito_client = mock_cognito
        cognito_provider._mfa_tokens = {"ChallengeName": "SMS_MFA", "Session": "test-session"}

        result = cognito_provider._complete_mfa_challenge("123456")

        assert result is True
        mock_cognito.respond_to_sms_mfa_challenge.assert_called_once_with(
            code="123456", mfa_tokens=cognito_provider._mfa_tokens
        )

    def test_complete_software_token_mfa_challenge(self, cognito_provider):
        """Test completing Software Token MFA challenge"""
        mock_cognito = Mock()
        cognito_provider.cognito_client = mock_cognito
        cognito_provider._mfa_tokens = {"ChallengeName": "SOFTWARE_TOKEN_MFA", "Session": "test-session"}

        result = cognito_provider._complete_mfa_challenge("123456")

        assert result is True
        mock_cognito.respond_to_software_token_mfa_challenge.assert_called_once_with(
            code="123456", mfa_tokens=cognito_provider._mfa_tokens
        )

    def test_complete_mfa_no_tokens(self, cognito_provider):
        """Test MFA challenge fails when no tokens"""
        result = cognito_provider._complete_mfa_challenge("123456")

        assert result is False

    def test_complete_mfa_challenge_error(self, cognito_provider):
        """Test MFA challenge handles errors"""
        mock_cognito = Mock()
        cognito_provider.cognito_client = mock_cognito
        cognito_provider._mfa_tokens = {"ChallengeName": "SMS_MFA"}
        mock_cognito.respond_to_sms_mfa_challenge.side_effect = Exception("Invalid code")

        result = cognito_provider._complete_mfa_challenge("999999")

        assert result is False


class TestCheckForMFAChallenges:
    """Test _check_for_mfa_challenges method"""

    @patch("amplify_excel_migrator.auth.cognito_auth.input")
    def test_sms_mfa_challenge(self, mock_input, cognito_provider):
        """Test handling SMS MFA challenge"""
        mock_admin_client = Mock()
        cognito_provider.boto_cognito_admin_client = mock_admin_client
        mock_input.return_value = "123456"

        response = {"ChallengeName": "SMS_MFA", "Session": "test-session"}

        cognito_provider._check_for_mfa_challenges(response, "test@example.com")

        mock_admin_client.admin_respond_to_auth_challenge.assert_called_once()

    @patch("amplify_excel_migrator.auth.cognito_auth.getpass")
    def test_new_password_required_challenge(self, mock_getpass, cognito_provider):
        """Test handling NEW_PASSWORD_REQUIRED challenge"""
        mock_admin_client = Mock()
        cognito_provider.boto_cognito_admin_client = mock_admin_client
        mock_getpass.return_value = "newpassword123"

        response = {"ChallengeName": "NEW_PASSWORD_REQUIRED", "Session": "test-session"}

        cognito_provider._check_for_mfa_challenges(response, "test@example.com")

        mock_admin_client.admin_respond_to_auth_challenge.assert_called_once()

    def test_mfa_setup_challenge(self, cognito_provider):
        """Test handling MFA_SETUP challenge"""
        response = {"ChallengeName": "MFA_SETUP"}

        result = cognito_provider._check_for_mfa_challenges(response, "test@example.com")

        assert result is False


class TestCheckUserInAdminsGroup:
    """Test _check_user_in_admins_group method"""

    @patch("amplify_excel_migrator.auth.cognito_auth.jwt.decode")
    def test_user_in_admin_group(self, mock_jwt_decode, cognito_provider):
        """Test user is in admin group"""
        mock_jwt_decode.return_value = {"cognito:groups": ["ADMINS", "USERS"]}

        cognito_provider._check_user_in_admins_group("test_token")

    @patch("amplify_excel_migrator.auth.cognito_auth.jwt.decode")
    def test_user_not_in_admin_group(self, mock_jwt_decode, cognito_provider):
        """Test user is not in admin group"""
        mock_jwt_decode.return_value = {"cognito:groups": ["USERS"]}

        with pytest.raises(PermissionError, match="User is not in ADMINS group"):
            cognito_provider._check_user_in_admins_group("test_token")

    @patch("amplify_excel_migrator.auth.cognito_auth.jwt.decode")
    def test_user_no_groups(self, mock_jwt_decode, cognito_provider):
        """Test user has no groups"""
        mock_jwt_decode.return_value = {}

        with pytest.raises(PermissionError, match="User is not in ADMINS group"):
            cognito_provider._check_user_in_admins_group("test_token")


class TestProviderInterface:
    """Test AuthenticationProvider interface implementation"""

    def test_get_id_token_not_authenticated(self, cognito_provider):
        """Test get_id_token returns None when not authenticated"""
        assert cognito_provider.get_id_token() is None

    def test_get_id_token_authenticated(self, cognito_provider, mock_id_token):
        """Test get_id_token returns token when authenticated"""
        cognito_provider._id_token = mock_id_token
        assert cognito_provider.get_id_token() == mock_id_token

    def test_is_authenticated_false(self, cognito_provider):
        """Test is_authenticated returns False when not authenticated"""
        assert cognito_provider.is_authenticated() is False

    def test_is_authenticated_true(self, cognito_provider, mock_id_token):
        """Test is_authenticated returns True when authenticated"""
        cognito_provider._id_token = mock_id_token
        assert cognito_provider.is_authenticated() is True



"""Tests for ConfigManager class"""

import json
import pytest
from pathlib import Path
from unittest.mock import patch, MagicMock
from amplify_excel_migrator.core import ConfigManager


@pytest.fixture
def sample_config():
    """Sample configuration for testing"""
    return {
        "excel_path": "test_data.xlsx",
        "api_endpoint": "https://test.appsync-api.us-east-1.amazonaws.com/graphql",
        "region": "us-east-1",
        "user_pool_id": "us-east-1_testpool",
        "client_id": "test-client-id",
        "username": "test@example.com",
    }


@pytest.fixture
def temp_config_manager(tmp_path):
    """Create ConfigManager with temporary config file"""
    config_file = tmp_path / "test_config.json"
    return ConfigManager(str(config_file))


class TestConfigManagerInitialization:
    """Test ConfigManager initialization"""

    def test_default_initialization(self):
        """Test ConfigManager with default config path"""
        manager = ConfigManager()
        assert manager.config_path == Path.home() / ".amplify-migrator" / "config.json"
        assert manager._config == {}

    def test_custom_path_initialization(self, tmp_path):
        """Test ConfigManager with custom config path"""
        custom_path = tmp_path / "custom_config.json"
        manager = ConfigManager(str(custom_path))
        assert manager.config_path == custom_path
        assert manager._config == {}


class TestConfigManagerLoad:
    """Test ConfigManager.load() method"""

    def test_load_existing_config(self, temp_config_manager, sample_config):
        """Test loading existing config file"""
        temp_config_manager.config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(temp_config_manager.config_path, "w") as f:
            json.dump(sample_config, f)

        result = temp_config_manager.load()
        assert result == sample_config
        assert temp_config_manager._config == sample_config

    def test_load_nonexistent_config(self, temp_config_manager):
        """Test loading when config file doesn't exist"""
        result = temp_config_manager.load()
        assert result == {}
        assert temp_config_manager._config == {}

    def test_load_corrupted_json(self, temp_config_manager):
        """Test loading corrupted JSON file"""
        temp_config_manager.config_path.parent.mkdir(parents=True, exist_ok=True)
        temp_config_manager.config_path.write_text("invalid json {")

        result = temp_config_manager.load()
        assert result == {}


class TestConfigManagerSave:
    """Test ConfigManager.save() method"""

    def test_save_config(self, temp_config_manager, sample_config):
        """Test saving configuration"""
        temp_config_manager.save(sample_config)

        assert temp_config_manager.config_path.exists()
        with open(temp_config_manager.config_path) as f:
            loaded = json.load(f)
            assert loaded == sample_config

    def test_save_creates_directory(self, tmp_path):
        """Test that save creates parent directories"""
        config_file = tmp_path / "nested" / "dir" / "config.json"
        manager = ConfigManager(str(config_file))

        assert not config_file.parent.exists()
        manager.save({"key": "value"})
        assert config_file.parent.exists()
        assert config_file.exists()

    def test_save_excludes_sensitive_keys(self, temp_config_manager):
        """Test that sensitive keys are not saved"""
        config_with_secrets = {
            "username": "test@example.com",
            "password": "secret123",
            "ADMIN_PASSWORD": "admin_secret",
            "api_endpoint": "https://api.example.com",
        }

        temp_config_manager.save(config_with_secrets)

        with open(temp_config_manager.config_path) as f:
            loaded = json.load(f)
            assert "password" not in loaded
            assert "ADMIN_PASSWORD" not in loaded
            assert "username" in loaded
            assert "api_endpoint" in loaded

    def test_save_updates_internal_config(self, temp_config_manager, sample_config):
        """Test that save updates the internal _config"""
        temp_config_manager.save(sample_config)
        # Note: internal _config excludes sensitive keys
        expected = {k: v for k, v in sample_config.items() if k not in ConfigManager.SENSITIVE_KEYS}
        assert temp_config_manager._config == expected


class TestConfigManagerGet:
    """Test ConfigManager.get() method"""

    def test_get_existing_key(self, temp_config_manager, sample_config):
        """Test getting existing key"""
        temp_config_manager._config = sample_config
        result = temp_config_manager.get("excel_path")
        assert result == "test_data.xlsx"

    def test_get_missing_key_with_default(self, temp_config_manager):
        """Test getting missing key with default value"""
        temp_config_manager._config = {}
        result = temp_config_manager.get("missing_key", "default_value")
        assert result == "default_value"

    def test_get_missing_key_without_default(self, temp_config_manager):
        """Test getting missing key without default"""
        temp_config_manager._config = {}
        result = temp_config_manager.get("missing_key")
        assert result is None

    def test_get_loads_config_if_empty(self, temp_config_manager, sample_config):
        """Test that get() loads config if _config is empty"""
        temp_config_manager.config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(temp_config_manager.config_path, "w") as f:
            json.dump(sample_config, f)

        temp_config_manager._config = {}
        result = temp_config_manager.get("excel_path")
        assert result == "test_data.xlsx"


class TestConfigManagerSet:
    """Test ConfigManager.set() method"""

    def test_set_new_key(self, temp_config_manager):
        """Test setting a new key"""
        temp_config_manager._config = {}
        temp_config_manager.set("new_key", "new_value")
        assert temp_config_manager._config["new_key"] == "new_value"

    def test_set_existing_key(self, temp_config_manager, sample_config):
        """Test updating an existing key"""
        temp_config_manager._config = sample_config.copy()
        temp_config_manager.set("excel_path", "new_path.xlsx")
        assert temp_config_manager._config["excel_path"] == "new_path.xlsx"

    def test_set_loads_config_if_empty(self, temp_config_manager, sample_config):
        """Test that set() loads config if _config is empty"""
        temp_config_manager.config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(temp_config_manager.config_path, "w") as f:
            json.dump(sample_config, f)

        temp_config_manager._config = {}
        temp_config_manager.set("new_key", "new_value")
        assert "excel_path" in temp_config_manager._config  # From loaded config
        assert temp_config_manager._config["new_key"] == "new_value"


class TestConfigManagerUpdate:
    """Test ConfigManager.update() method"""

    def test_update_config(self, temp_config_manager, sample_config):
        """Test updating config with multiple values"""
        temp_config_manager._config = sample_config.copy()
        updates = {"excel_path": "updated.xlsx", "new_key": "new_value"}

        temp_config_manager.update(updates)

        assert temp_config_manager._config["excel_path"] == "updated.xlsx"
        assert temp_config_manager._config["new_key"] == "new_value"
        assert temp_config_manager.config_path.exists()

    def test_update_saves_to_file(self, temp_config_manager):
        """Test that update() persists changes to file"""
        temp_config_manager._config = {"old_key": "old_value"}
        temp_config_manager.update({"new_key": "new_value"})

        with open(temp_config_manager.config_path) as f:
            loaded = json.load(f)
            assert loaded["old_key"] == "old_value"
            assert loaded["new_key"] == "new_value"


class TestConfigManagerPromptForValue:
    """Test ConfigManager.prompt_for_value() method"""

    def test_prompt_with_default_empty_input(self, temp_config_manager):
        """Test prompt with default when user provides empty input"""
        with patch("builtins.input", return_value=""):
            result = temp_config_manager.prompt_for_value("Test prompt", "default_value")
            assert result == "default_value"

    def test_prompt_with_user_input(self, temp_config_manager):
        """Test prompt with user input"""
        with patch("builtins.input", return_value="user_value"):
            result = temp_config_manager.prompt_for_value("Test prompt", "default_value")
            assert result == "user_value"

    def test_prompt_without_default(self, temp_config_manager):
        """Test prompt without default"""
        with patch("builtins.input", return_value="custom_value"):
            result = temp_config_manager.prompt_for_value("Test prompt")
            assert result == "custom_value"

    def test_prompt_secret_input(self, temp_config_manager):
        """Test prompting for secret (password) input"""
        with patch("amplify_excel_migrator.core.config.getpass", return_value="secret123"):
            result = temp_config_manager.prompt_for_value("Password", secret=True)
            assert result == "secret123"

    def test_prompt_strips_whitespace(self, temp_config_manager):
        """Test that input is stripped of whitespace"""
        with patch("builtins.input", return_value="  value with spaces  "):
            result = temp_config_manager.prompt_for_value("Test prompt")
            assert result == "value with spaces"

    def test_prompt_displays_default_in_brackets(self, temp_config_manager):
        """Test that default is shown in brackets in prompt"""
        with patch("builtins.input", return_value="") as mock_input:
            temp_config_manager.prompt_for_value("Test prompt", "default123")
            mock_input.assert_called_once_with("Test prompt [default123]: ")


class TestConfigManagerGetOrPrompt:
    """Test ConfigManager.get_or_prompt() method"""

    def test_get_or_prompt_returns_cached(self, temp_config_manager, sample_config):
        """Test that cached value is returned without prompting"""
        temp_config_manager._config = sample_config
        with patch("builtins.input", return_value="should_not_be_used"):
            result = temp_config_manager.get_or_prompt("excel_path", "Excel path")
            assert result == "test_data.xlsx"

    def test_get_or_prompt_prompts_when_missing(self, temp_config_manager):
        """Test that prompt is shown when key is missing"""
        temp_config_manager._config = {}
        with patch("builtins.input", return_value="new_value"):
            result = temp_config_manager.get_or_prompt("missing_key", "Test prompt")
            assert result == "new_value"

    def test_get_or_prompt_with_default(self, temp_config_manager):
        """Test get_or_prompt with default value"""
        temp_config_manager._config = {}
        with patch("builtins.input", return_value=""):
            result = temp_config_manager.get_or_prompt("missing_key", "Test prompt", "default123")
            assert result == "default123"

    def test_get_or_prompt_loads_config_if_empty(self, temp_config_manager, sample_config):
        """Test that get_or_prompt loads config if _config is empty"""
        temp_config_manager.config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(temp_config_manager.config_path, "w") as f:
            json.dump(sample_config, f)

        temp_config_manager._config = {}
        result = temp_config_manager.get_or_prompt("excel_path", "Excel path")
        assert result == "test_data.xlsx"


class TestConfigManagerExists:
    """Test ConfigManager.exists() method"""

    def test_exists_when_file_exists(self, temp_config_manager):
        """Test exists() returns True when config file exists"""
        temp_config_manager.config_path.parent.mkdir(parents=True, exist_ok=True)
        temp_config_manager.config_path.touch()
        assert temp_config_manager.exists() is True

    def test_exists_when_file_not_exists(self, temp_config_manager):
        """Test exists() returns False when config file doesn't exist"""
        assert temp_config_manager.exists() is False


class TestConfigManagerClear:
    """Test ConfigManager.clear() method"""

    def test_clear_removes_file(self, temp_config_manager, sample_config):
        """Test that clear removes the config file"""
        temp_config_manager.save(sample_config)
        assert temp_config_manager.config_path.exists()

        temp_config_manager.clear()
        assert not temp_config_manager.config_path.exists()

    def test_clear_resets_internal_config(self, temp_config_manager, sample_config):
        """Test that clear resets internal _config"""
        temp_config_manager._config = sample_config
        temp_config_manager.clear()
        assert temp_config_manager._config == {}

    def test_clear_when_file_not_exists(self, temp_config_manager):
        """Test that clear doesn't fail when file doesn't exist"""
        assert not temp_config_manager.config_path.exists()
        temp_config_manager.clear()  # Should not raise
        assert temp_config_manager._config == {}


class TestConfigManagerIntegration:
    """Integration tests for ConfigManager"""

    def test_full_workflow(self, temp_config_manager, sample_config):
        """Test complete workflow: save, load, update, clear"""
        temp_config_manager.save(sample_config)
        assert temp_config_manager.exists()

        loaded = temp_config_manager.load()
        assert loaded["excel_path"] == "test_data.xlsx"

        temp_config_manager.update({"excel_path": "updated.xlsx"})
        assert temp_config_manager.get("excel_path") == "updated.xlsx"

        temp_config_manager.clear()
        assert not temp_config_manager.exists()
        assert temp_config_manager._config == {}

    def test_multiple_managers_same_file(self, tmp_path, sample_config):
        """Test multiple ConfigManager instances with same file"""
        config_file = tmp_path / "shared_config.json"

        manager1 = ConfigManager(str(config_file))
        manager1.save(sample_config)

        manager2 = ConfigManager(str(config_file))
        loaded = manager2.load()
        assert loaded == sample_config



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="amplify-excel-migrator",
    version="1.2.8",
    author="Eyal Politansky",
    author_email="10eyal10@gmail.com",
    description="A CLI tool to migrate Excel data to AWS Amplify",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EyalPoly/amplify-excel-migrator",
    py_modules=["migrator", "amplify_client", "model_field_parser"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    install_requires=[
        "pandas>=1.3.0",
        "requests>=2.26.0",
        "boto3>=1.18.0",
        "pycognito>=2023.5.0",
        "PyJWT>=2.0.0",
        "aiohttp>=3.8.0",
        "openpyxl>=3.0.0",
        "inflect>=7.0.0",
    ],
    extras_require={
        "dev": [
            "pytest>=8.0.0",
            "pytest-cov>=4.1.0",
            "pytest-mock>=3.12.0",
            "pytest-asyncio>=0.21.0",
            "setuptools>=80.0.0",
            "wheel>=0.40.0",
            "twine>=4.0.0",
        ],
    },
    entry_points={
        "console_scripts": [
            "amplify-migrator=migrator:main",
        ],
    },
)
